from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.action.action_resolver import ActionResolver
from hackle.internal.evaluation.container.container_resolver import ContainerResolver
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.flow.evaluation_flow import EvaluationFlow
from hackle.internal.evaluation.flow.flow_evaluator import OverrideEvaluator, DraftEvaluator, PausedEvaluator, \
    CompletedEvaluator, ExperimentTargetEvaluator, TrafficAllocateEvaluator, TargetRuleEvaluator, DefaultRuleEvaluator, \
    ContainerEvaluator, IdentifierEvaluator
from hackle.internal.evaluation.target.experiment_target_determiner import ExperimentTargetDeterminer
from hackle.internal.evaluation.target.experiment_target_rule_determiner import ExperimentTargetRuleDeterminer
from hackle.internal.evaluation.target.override_resolver import OverrideResolver
from hackle.internal.model.entities import TargetRule
from hackle.internal.workspace.workspace import Workspace
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment, create_variation


class FlowEvaluatorTest(TestCase):

    def setUp(self):
        self.evaluation = Mock(spec=ExperimentEvaluation)
        self.next_flow = Mock(spec=EvaluationFlow)
        self.next_flow.evaluate.return_value = self.evaluation
        self.context = Evaluator.context()


class OverrideEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.override_resolver = Mock(spec=OverrideResolver)
        self.sut = OverrideEvaluator(self.override_resolver)

    def test_unsupported_type(self):
        experiment = create_experiment(type='INVALID')
        request = create_experiment_request(experiment=experiment)

        with self.assertRaises(Exception):
            self.sut.evaluate(request, self.context, self.next_flow)

    def test_AB_TEST에_수동할당된_사용자인_경우_수동할당된_그룹으로_평가한다(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        variation = experiment.variations[0]
        self.override_resolver.resolve_or_none.return_value = variation

        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.OVERRIDDEN, actual.reason)
        self.assertEqual(variation.id, actual.variation_id)

    def test_FEATURE_FLAG에_개별타겟된_사용자는_개별타겟된_그룹으로_평가한다(self):
        # given
        experiment = create_experiment(type='FEATURE_FLAG')
        variation = experiment.variations[0]
        self.override_resolver.resolve_or_none.return_value = variation

        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.INDIVIDUAL_TARGET_MATCH, actual.reason)
        self.assertEqual(variation.id, actual.variation_id)

    def test_수동할당된_사용자가_아닌경우_다음_flow로_평가한다(self):
        # given
        self.override_resolver.resolve_or_none.return_value = None

        # when
        actual = self.sut.evaluate(create_experiment_request(), self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()


class DraftEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.sut = DraftEvaluator()

    def test_DRAFT_상태면_기본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(
            type='AB_TEST',
            status='DRAFT',
            variations=[
                create_variation(id=42, key='A'),
                create_variation(id=43, key='B')
            ]
        )
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.EXPERIMENT_DRAFT, actual.reason)
        self.assertEqual(42, actual.variation_id)

    def test_DRAFT_상태가_아니면_다름_플로우로_평가한다(self):
        # given
        experiment = create_experiment(type='AB_TEST', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()


class PausedEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.sut = PausedEvaluator()

    def test_unsupported_type(self):
        experiment = create_experiment(type='INVALID', status='PAUSED')
        request = create_experiment_request(experiment=experiment)

        with self.assertRaises(Exception):
            self.sut.evaluate(request, self.context, self.next_flow)

    def test_AB_TEST의_상태가_PAUSED면_기본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(
            type='AB_TEST',
            status='PAUSED',
            variations=[
                create_variation(id=42, key='A'),
                create_variation(id=43, key='B')
            ]
        )
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.EXPERIMENT_PAUSED, actual.reason)
        self.assertEqual(42, actual.variation_id)

    def test_FEATURE_FLAG의_상태가_PAUSED면_기본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(
            type='FEATURE_FLAG',
            status='PAUSED',
            variations=[
                create_variation(id=42, key='A'),
                create_variation(id=43, key='B')
            ]
        )
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.FEATURE_FLAG_INACTIVE, actual.reason)
        self.assertEqual(42, actual.variation_id)

    def test_PAUSED_상태가_아니면_다름_플로우로_평가한다(self):
        # given
        experiment = create_experiment(type='AB_TEST', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()


class CompletedEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.sut = CompletedEvaluator()

    def test_COMPLETED_상태면_기본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(
            type='AB_TEST',
            status='COMPLETED',
            variations=[
                create_variation(id=42, key='A'),
                create_variation(id=43, key='B')
            ],
            winner_variation_id=43
        )
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.EXPERIMENT_COMPLETED, actual.reason)
        self.assertEqual(43, actual.variation_id)

    def test_COMPLETED_상태지만_winner가_없으면_예외(self):
        # given
        experiment = create_experiment(
            id=320,
            type='AB_TEST',
            status='COMPLETED',
            variations=[
                create_variation(id=42, key='A'),
                create_variation(id=43, key='B')
            ]
        )
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('Winner variation[320]', str(actual.exception))

    def test_COMPLETED_상태가_아니면_다름_플로우로_평가한다(self):
        # given
        experiment = create_experiment(type='AB_TEST', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()


class ExperimentTargetEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.experiment_target_determiner = Mock(spec=ExperimentTargetDeterminer)
        self.sut = ExperimentTargetEvaluator(self.experiment_target_determiner)

    def test_AB_TEST_타입이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment type must be AB_TEST [42]', str(actual.exception))

    def test_사용자가_실험_참여_대상이면_다음_플로우로_평가한다(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = create_experiment_request(experiment=experiment)

        self.experiment_target_determiner.is_user_in_experiment_target.return_value = True

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()

    def test_사용자가_실험_참여_대상이_아니면_기본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = create_experiment_request(experiment=experiment)

        self.experiment_target_determiner.is_user_in_experiment_target.return_value = False

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.NOT_IN_EXPERIMENT_TARGET, actual.reason)
        self.assertEqual('A', actual.variation_key)


class TrafficAllocateEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.action_resolver = Mock(spec=ActionResolver)
        self.sut = TrafficAllocateEvaluator(self.action_resolver)

    def test_실행중이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='AB_TEST', status='DRAFT')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment status must be RUNNING [42]', str(actual.exception))

    def test_AB_TEST_타입이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment type must be AB_TEST [42]', str(actual.exception))

    def test_기본룰에_해당하는_Variation이_없으면_기본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(id=42, type='AB_TEST', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        self.action_resolver.resolve_or_none.return_value = None

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.TRAFFIC_NOT_ALLOCATED, actual.reason)
        self.assertEqual('A', actual.variation_key)

    def test_할당된_그룹이_드랍되었으면_그본그룹으로_평가한다(self):
        # given
        experiment = create_experiment(
            id=42, type='AB_TEST', status='RUNNING',
            variations=[
                create_variation(id=41, key='A'),
                create_variation(id=42, key='B'),
                create_variation(id=43, key='C', is_dropped=True),
            ]
        )
        request = create_experiment_request(experiment=experiment)

        self.action_resolver.resolve_or_none.return_value = experiment.get_variation_by_key_or_none('C')

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.VARIATION_DROPPED, actual.reason)
        self.assertEqual('A', actual.variation_key)

    def test_할당된_그룹으로_평가한다(self):
        # given
        experiment = create_experiment(
            id=42, type='AB_TEST', status='RUNNING',
            variations=[
                create_variation(id=41, key='A'),
                create_variation(id=42, key='B'),
            ]
        )
        request = create_experiment_request(experiment=experiment)

        self.action_resolver.resolve_or_none.return_value = experiment.get_variation_by_key_or_none('B')

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, actual.reason)
        self.assertEqual(42, actual.variation_id)


class TargetRuleEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.target_rule_determiner = Mock(spec=ExperimentTargetRuleDeterminer)
        self.action_resolver = Mock(spec=ActionResolver)
        self.sut = TargetRuleEvaluator(self.target_rule_determiner, self.action_resolver)

    def test_실행중이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='DRAFT')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment status must be RUNNING [42]', str(actual.exception))

    def test_FEATURE_FLAG_타입이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='AB_TEST', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment type must be FEATURE_FLAG [42]', str(actual.exception))

    def test_identifier_type에_해당하는_식별자가_없으면_다음_플로우를_실행한다(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING', identifier_type='custom_id')
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()

    def test_타겟룰에_해당하지_않으면_다음_플로우로_평가한다(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        self.target_rule_determiner.determine_target_rule_or_none.return_value = None

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()

    def test_타겟룰에_해당하지만_Action에_해당하는_Variation이_결정되지_않으면_예외발생(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING')
        request = create_experiment_request(experiment=experiment)
        target_rule = TargetRule(Mock(), Mock())

        self.target_rule_determiner.determine_target_rule_or_none.return_value = target_rule
        self.action_resolver.resolve_or_none.return_value = None

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('FeatureFlag must decide the variation [42]', str(actual.exception))
        self.target_rule_determiner.determine_target_rule_or_none.assert_called_once()
        self.action_resolver.resolve_or_none.assert_called_once()

    def test_타겟룰에_해당하면_해당룰로_결정된_Variation으로_평가한다(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING')
        request = create_experiment_request(experiment=experiment)
        target_rule = TargetRule(Mock(), Mock())

        self.target_rule_determiner.determine_target_rule_or_none.return_value = target_rule
        self.action_resolver.resolve_or_none.return_value = experiment.variations[0]

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.TARGET_RULE_MATCH, actual.reason)
        self.assertEqual(experiment.variations[0].id, actual.variation_id)


class DefaultRuleEvaluatorTest(FlowEvaluatorTest):
    def setUp(self):
        super().setUp()
        self.action_resolver = Mock(spec=ActionResolver)
        self.sut = DefaultRuleEvaluator(self.action_resolver)

    def test_실행중이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='DRAFT')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment status must be RUNNING [42]', str(actual.exception))

    def test_FEATURE_FLAG_타입이_아니면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='AB_TEST', status='RUNNING')
        request = create_experiment_request(experiment=experiment)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('experiment type must be FEATURE_FLAG [42]', str(actual.exception))

    def test_identifier_type에_해당하는_식별자가_없으면_default_variation을_리턴한다(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING', identifier_type='custom_id')
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.DEFAULT_RULE, actual.reason)
        self.assertEqual('A', actual.variation_key)

    def test_기본룰에_해당하는_Variation을_결정하지_못하면_예외_발생(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING')
        request = create_experiment_request(experiment=experiment)
        self.action_resolver.resolve_or_none.return_value = None

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('FeatureFlag must decide the variation [42]', str(actual.exception))

    def test_기본룰에_해당하는_Variation으로_평가한다(self):
        # given
        experiment = create_experiment(id=42, type='FEATURE_FLAG', status='RUNNING')
        request = create_experiment_request(experiment=experiment)
        self.action_resolver.resolve_or_none.return_value = experiment.variations[0]

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.DEFAULT_RULE, actual.reason)
        self.assertEqual(experiment.variations[0].id, actual.variation_id)


class ContainerEvaluatorTest(FlowEvaluatorTest):

    def setUp(self):
        super().setUp()
        self.container_resolver = Mock(spec=ContainerResolver)
        self.sut = ContainerEvaluator(self.container_resolver)

    def test_상호배타적실험이_아니면_다음_플로우_실행(self):
        # given
        workspace = Mock(spec=Workspace)
        experiment = create_experiment()
        request = create_experiment_request(workspace=workspace, experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()

    def test_container_id로_Container를_찾을수_없으면_예외발생(self):
        # given
        workspace = Mock(spec=Workspace)
        experiment = create_experiment(id=42, container_id=320)
        request = create_experiment_request(workspace=workspace, experiment=experiment)

        workspace.get_container_or_none.return_value = None

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual('Container[320]', str(actual.exception))

    def test_상호배타그룹이면_다음플로우(self):
        # given
        workspace = Mock(spec=Workspace)
        experiment = create_experiment(id=42, container_id=320)
        request = create_experiment_request(workspace=workspace, experiment=experiment)

        workspace.get_container_or_none.return_value = Mock()
        self.container_resolver.is_user_in_container_group.return_value = True

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()

    def test_상호배타에_속해있지만_상호배타그룹에_포함하지_않으면_기본그룹(self):
        # given
        workspace = Mock(spec=Workspace)
        experiment = create_experiment(id=42, container_id=320)
        request = create_experiment_request(workspace=workspace, experiment=experiment)

        workspace.get_container_or_none.return_value = Mock()
        self.container_resolver.is_user_in_container_group.return_value = False

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT, actual.reason)
        self.assertEqual('A', actual.variation_key)


class IdentifierEvaluatorTest(FlowEvaluatorTest):
    def setUp(self):
        super().setUp()
        self.sut = IdentifierEvaluator()

    def test_식별자가_있으면_다음플로우_실행(self):
        # given
        experiment = create_experiment()
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(self.evaluation, actual)
        self.next_flow.evaluate.assert_called_once()

    def test_식별자가_없으면_IDENTIFIER_NOT_FOUND(self):
        experiment = create_experiment(identifier_type='custom_id')
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.evaluate(request, self.context, self.next_flow)

        # then
        self.assertEqual(DecisionReason.IDENTIFIER_NOT_FOUND, actual.reason)
        self.assertEqual('A', actual.variation_key)



import json

from hackle.internal.model.entities import *


class Workspace(object):
    def __init__(self, data):
        json_data = json.loads(data)

        self.bucket_id_map = self._bucket_id_map(json_data.get('buckets', []))
        self.experiment_key_map = self._experiment_key_map('AB_TEST', json_data.get('experiments', []))
        self.feature_flag_key_map = self._experiment_key_map('FEATURE_FLAG', json_data.get('featureFlags', []))
        self.event_type_key_map = self._event_type_key_map(json_data.get('events', []))
        self.segment_key_map = self._segment_key_map(json_data.get('segments', []))
        self._container_id_map = self._container_id_map(json_data.get('containers', []))
        self.parameter_configurations = self._parameter_configuration_id_map(
            json_data.get('parameterConfigurations', []))
        self.remote_config_parameters = self._remote_config_parameter_key_map(
            json_data.get('remoteConfigParameters', []))

    def get_experiment_or_none(self, experiment_key):
        return self.experiment_key_map.get(experiment_key)

    def get_feature_flag_or_none(self, feature_key):
        return self.feature_flag_key_map.get(feature_key)

    def get_bucket_or_none(self, bucket_id):
        return self.bucket_id_map.get(bucket_id)

    def get_event_type_or_none(self, event_key):
        return self.event_type_key_map.get(event_key)

    def get_segment_or_none(self, segment_key):
        return self.segment_key_map.get(segment_key)

    def get_container_or_none(self, container_id):
        return self._container_id_map.get(container_id)

    def get_parameter_configuration_or_none(self, parameter_configuration_id):
        return self.parameter_configurations.get(parameter_configuration_id)

    def get_remote_config_parameter_or_none(self, remote_config_parameter_key):
        return self.remote_config_parameters.get(remote_config_parameter_key)

    @staticmethod
    def _bucket_id_map(buckets_data):
        bucket_id_map = {}
        for bucket_data in buckets_data:
            slots = []

            for slot_data in bucket_data['slots']:
                slots.append(
                    Slot(
                        start_inclusive=slot_data['startInclusive'],
                        end_exclusive=slot_data['endExclusive'],
                        variation_id=slot_data['variationId']
                    )
                )

            bucket_id_map[bucket_data['id']] = Bucket(
                seed=bucket_data['seed'],
                slot_size=bucket_data['slotSize'],
                slots=slots
            )
        return bucket_id_map

    @staticmethod
    def _experiment_key_map(experiment_type, experiments_data):
        experiment_key_map = {}
        for experiment_data in experiments_data:
            experiment = Workspace._experiment(experiment_type, experiment_data)
            if experiment:
                experiment_key_map[experiment_data['key']] = experiment
        return experiment_key_map

    @staticmethod
    def _experiment(type, experiment_data):

        execution_data = experiment_data['execution']
        status = Workspace._experiment_status_or_none(execution_data['status'])
        if not status:
            return None

        variations = []
        for variation_data in experiment_data['variations']:
            variation = Variation(
                id=variation_data['id'],
                key=variation_data['key'],
                is_dropped=variation_data['status'] == 'DROPPED',
                parameter_configuration_id=variation_data.get('parameterConfigurationId')
            )
            variations.append(variation)

        user_overrides = {}
        for override_data in execution_data['userOverrides']:
            user_overrides[override_data['userId']] = override_data['variationId']

        return Experiment(
            id=experiment_data['id'],
            key=experiment_data['key'],
            type=type,
            identifier_type=experiment_data['identifierType'],
            status=status,
            version=experiment_data['version'],
            variations=variations,
            user_overrides=user_overrides,
            segment_overrides=Workspace._target_rules(execution_data.get('segmentOverrides', []),
                                                      TargetingTypes.IDENTIFIER),
            target_audiences=Workspace._targets(execution_data.get('targetAudiences', []), TargetingTypes.PROPERTY),
            target_rules=Workspace._target_rules(execution_data.get('targetRules', []), TargetingTypes.PROPERTY),
            default_rule=Workspace._target_action(execution_data['defaultRule']),
            container_id=experiment_data.get('containerId'),
            winner_variation_id=experiment_data.get('winnerVariationId')
        )

    @staticmethod
    def _experiment_status_or_none(execution_status):
        if execution_status == 'READY':
            return 'DRAFT'
        elif execution_status == 'RUNNING':
            return 'RUNNING'
        elif execution_status == 'PAUSED':
            return 'PAUSED'
        elif execution_status == 'STOPPED':
            return 'COMPLETED'
        else:
            return None

    @staticmethod
    def _target_or_none(target_data, targeting_type):
        conditions = []

        for condition_data in target_data['conditions']:
            condition = Workspace._condition_or_none(condition_data, targeting_type)
            if condition:
                conditions.append(condition)

        if not conditions:
            return None

        return Target(conditions)

    @staticmethod
    def _condition_or_none(condition_data, targeting_type):

        target_key = Workspace._target_key(condition_data['key'])

        if not targeting_type.supports(target_key.type):
            return None

        target_match = Workspace._target_match(condition_data['match'])

        return TargetCondition(
            key=target_key,
            match=target_match
        )

    @staticmethod
    def _target_key(target_key_data):
        return TargetKey(
            type=target_key_data['type'],
            name=target_key_data['name']
        )

    @staticmethod
    def _target_match(target_match_data):
        return TargetMatch(
            type=target_match_data['type'],
            operator=target_match_data['operator'],
            value_type=target_match_data['valueType'],
            values=target_match_data['values']
        )

    @staticmethod
    def _target_action(target_action_data):
        return TargetAction(
            type=target_action_data['type'],
            variation_id=target_action_data.get('variationId'),
            bucket_id=target_action_data.get('bucketId')
        )

    @staticmethod
    def _target_rule_or_none(target_rule_data, targeting_type):
        target = Workspace._target_or_none(target_rule_data['target'], targeting_type)
        if not target:
            return None

        action = Workspace._target_action(target_rule_data['action'])

        return TargetRule(
            target=target,
            action=action
        )

    @staticmethod
    def _targets(targets_data, targeting_type):
        targets = []
        for target_data in targets_data:
            target = Workspace._target_or_none(target_data, targeting_type)
            if not target:
                continue
            targets.append(target)
        return targets

    @staticmethod
    def _target_rules(target_rules_data, targeting_type):
        target_rules = []
        for target_rule_data in target_rules_data:
            target_rule = Workspace._target_rule_or_none(target_rule_data, targeting_type)
            if not target_rule:
                continue
            target_rules.append(target_rule)
        return target_rules

    @staticmethod
    def _event_type_key_map(event_types_data):
        event_type_key_map = {}
        for event_type_data in event_types_data:
            event_type_key_map[str(event_type_data['key'])] = EventType(event_type_data['id'],
                                                                        event_type_data['key'])
        return event_type_key_map

    @staticmethod
    def _segment_key_map(segments_data):
        _segment_key_map = {}
        for segment_data in segments_data:
            segment = Workspace._segment(segment_data)
            _segment_key_map[segment.key] = segment
        return _segment_key_map

    @staticmethod
    def _segment(segment_data):
        return Segment(
            id=segment_data['id'],
            key=segment_data['key'],
            type=segment_data['type'],
            targets=Workspace._targets(segment_data['targets'], TargetingTypes.SEGMENT)
        )

    @staticmethod
    def _container_id_map(containers_data):
        _container_id_map = {}
        for container_data in containers_data:
            container = Workspace._container(container_data)
            _container_id_map[container.id] = container
        return _container_id_map

    @staticmethod
    def _container(container_data):
        groups = []
        for container_group_data in container_data.get('groups', []):
            container_group = Workspace._container_group(container_group_data)
            groups.append(container_group)
        return Container(
            id=container_data['id'],
            bucket_id=container_data['bucketId'],
            groups=groups
        )

    @staticmethod
    def _container_group(container_group_data):
        return ContainerGroup(
            id=container_group_data['id'],
            experiments=container_group_data['experiments']
        )

    @staticmethod
    def _parameter_configuration_id_map(parameter_configurations_data):
        _parameter_configuration_id_map = {}
        for parameter_configuration_data in parameter_configurations_data:
            parameters = {}
            for parameter in parameter_configuration_data['parameters']:
                parameters[parameter['key']] = parameter['value']

            parameter_configuration = ParameterConfiguration(
                id=parameter_configuration_data['id'],
                parameters=parameters
            )
            _parameter_configuration_id_map[parameter_configuration.id] = parameter_configuration
        return _parameter_configuration_id_map

    @staticmethod
    def _remote_config_parameter_key_map(remote_config_parameters_data):
        _remote_config_parameter_key_map = {}
        for remote_config_parameter_data in remote_config_parameters_data:
            target_rules = []
            for target_rule_data in remote_config_parameter_data['targetRules']:
                target_rules.append(Workspace._remote_config_parameter_target_rule_or_none(target_rule_data, TargetingTypes.PROPERTY))

            default_value = Workspace._remote_config_parameter_value(remote_config_parameter_data['defaultValue'])

            remote_config_parameter = RemoteConfigParameter(
                id=remote_config_parameter_data['id'],
                key=remote_config_parameter_data['key'],
                type=remote_config_parameter_data['type'],
                identifier_type=remote_config_parameter_data['identifierType'],
                target_rules=target_rules,
                default_value=default_value
            )

            _remote_config_parameter_key_map[remote_config_parameter_data['key']] = remote_config_parameter
        return _remote_config_parameter_key_map

    @staticmethod
    def _remote_config_parameter_target_rule_or_none(target_rule_data, targeting_type):
        target = Workspace._target_or_none(target_rule_data['target'], targeting_type)
        if not target:
            return None

        return RemoteConfigTargetRule(
            key=target_rule_data['key'],
            name=target_rule_data['name'],
            target=target,
            bucket_id=target_rule_data['bucketId'],
            value=Workspace._remote_config_parameter_value(target_rule_data['value'])
        )

    @staticmethod
    def _remote_config_parameter_value(remote_config_parameter_value_data):
        return RemoteConfigParameterValue(
            id=remote_config_parameter_value_data['id'],
            raw_value=remote_config_parameter_value_data['value']
        )
